#include <stdio.h>
#include <string.h>
#include "iodefine.h"
#include "cprintf.h"
#include "timer.h"
#include "osboard.h"
#include <fatfs.h>

#include "mad.h"

/* Set this to 1 to enable timer buckets; each bucket B_* stores uS
   values for that phase of the processing.  */
#define ENABLE_BUCKETS 0

static enum {
  B_idle,
  B_input,
  B_disk,
  B_processing,
  B_output,
  B_output_p,
  B_wait,
  B_printf
} BUCKET;

#if ENABLE_BUCKETS

static long buckets[10];
static int last_bucket = B_idle;
static long last_bucket_time;

static void
bucket (int which)
{
  unsigned short mtu0 = MTU0.TCNT;
  unsigned short ms = timer_ms;
  long count = (long)timer_ms * 1000 + (long)mtu0 / 12;
  buckets[last_bucket] = count - last_bucket_time;
  last_bucket = which;
  last_bucket_time = count;
  LED = 1 << which;
}
#else
#define bucket(which)  LED = 1 << which
#endif

static unsigned char ibuf[1440 + 512];

typedef struct {
  /* pointers to our parent's buffers/files/etc */
  unsigned char const *start;
  unsigned int length;
  /* our RAM sample buffer */
  unsigned char const *ibuf;
  unsigned int ilength;
  unsigned int isize;
} Buffer;

/* Largest single-frame sample size.  */
#define DAC_BUF_SIZE 1152

/* A buffer is empty if LENGTH is zero.  If LENGTH is nonzero and
   equal to PTR, the buffer is full but idle.  If PTR is less than
   LENGTH, the buffer is actively streaming to the DAC.  */
typedef struct {
  int length;
  int ptr;
  int sample_rate;
  unsigned short left[DAC_BUF_SIZE];
  unsigned short right[DAC_BUF_SIZE];
} DAC_Buf;

/* Note that TWO volatiles are needed for the pointers, one for the
   pointer itself (they get swapped inside the ISR) and one for the
   data they point to (ptr++)  */
static volatile DAC_Buf dac_buf[2];
static volatile DAC_Buf * volatile dac_active;
static volatile DAC_Buf * volatile dac_idle;
static int mtu2_sample_rate = 0;

/*----------------------------------------------------------------------*/

/* This handler happens 44100 times per second, or whatever the MP3
   sample rate is.  The common paths are: streaming from a buffer
   (first IF), and waiting for any data (second IF).  It only drops
   all the way through when it's switching buffers.  */

static void __attribute__((interrupt))
mtu2_handler ()
{
  volatile DAC_Buf *tmp;

  /* If the active buffer still has data, stream it.  */
  if (dac_active->ptr < dac_active->length)
    {
      /* Here is where you do "something" with the two samples.  In
	 this example, we're feeding them to the two DACs, but you can
	 also send them to a PWM or external DAC, or just store them
	 somewhere.  */
      DA.DADR0 = dac_active->left[dac_active->ptr];
      DA.DADR1 = dac_active->right[dac_active->ptr];

      dac_active->ptr ++;
      /* If there's still data in this buffer, we're done.  */
      if (dac_active->ptr < dac_active->length)
	return;
    }

  /* If the other buffer is also empty, both buffers are now empty,
     and we're idle.  */
  if (dac_idle->ptr >= dac_idle->length)
    return;

  /* The other buffer has data.  Switch buffers.  */
  tmp = dac_active;
  dac_active = dac_idle;
  dac_idle = tmp;
  dac_idle->length = dac_idle->ptr = 0;

  /* update MTU2 rate based on new sample rate? */

  if (mtu2_sample_rate != dac_active->sample_rate)
    {
      mtu2_sample_rate = dac_active->sample_rate;
      MTU2.TGRA = 48000000 / dac_active->sample_rate;
    }
}

/*----------------------------------------------------------------------*/

/* This INPUT function is for when the data is stored in ROM (flash).
   We simply feed the whole MP3 file to the decoder at once.  */

static
enum mad_flow input_rom(void *data,
			struct mad_stream *stream)
{
  Buffer *buffer = (Buffer *)data;

  if (buffer->length <= 0)
    return MAD_FLOW_STOP;

  iprintf("mp3: input %d bytes\n", buffer->length);

  mad_stream_buffer (stream, buffer->start, buffer->length);
  buffer->length = 0;

  return MAD_FLOW_CONTINUE;
}

/*----------------------------------------------------------------------*/

/* This INPUT function is for file-based (or other off-chip stream)
   MP3s.  Each time this is called, the "read_file" function (replace
   it with whatever is appropriate to you) should read as much as it
   can, up to new_bytes, but it's OK to read slightly less (i.e. if it
   wants to read in whole blocks of 512 bytes, for example)  */

static
enum mad_flow input_file(void *data,
			struct mad_stream *stream)
{
  Buffer *buffer = (Buffer *)data;
  int i;
  int old_bytes = 0;
  int max_bytes;
  int new_bytes;
  unsigned char *new_ptr;

  i = buffer->length;

  if (stream->this_frame && stream->next_frame)
    {
      old_bytes = buffer->ilength - (unsigned int)(stream->next_frame - stream->buffer);
      memmove((void *)stream->buffer, (void *)stream->next_frame, old_bytes);
      new_bytes = buffer->isize - old_bytes;
      new_ptr = buffer->ibuf + old_bytes;
    }
  else
    {
      old_bytes = 0;
      new_bytes = buffer->isize;
      new_ptr = buffer->ibuf;
    }

  bucket(B_disk);
  new_bytes = read_file (new_ptr, new_bytes);
  bucket(B_input);
  if (new_bytes <= 0)
    {
      iprintf("EOF\n");
      return MAD_FLOW_STOP;
    }

  buffer->ilength = old_bytes + new_bytes;

  mad_stream_buffer (stream, buffer->ibuf, old_bytes + new_bytes);

  bucket(B_processing);
  return MAD_FLOW_CONTINUE;
}

/*----------------------------------------------------------------------*/

/* This is the MP3 output function.  All it does is convert the
   samples to our format (16 bit unsigned) and store them in the next
   available buffer.  Note that this is where most of the time is
   spent - waiting for the next buffer.  */

static
enum mad_flow output(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  int i;
  Buffer *buffer = (Buffer *)data;
  short wait_start, wait_end;
  static unsigned short wait_time;

  /* Wait for the idle buffer to be empty.  */
  bucket(B_wait);
  while (dac_idle->length > 0)
    {
      /* If you have something to do while waiting for a buffer, this
	 is the place to do it.  */
      asm("wait");
    }
  bucket(B_output);

  /* Convert the 4+28 signed fixed point samples to our format, 16-bit
     unsigned.  */
#define sample2DA(x) (((int)(x) >> 12) + 0x8000)
  for (i=0; i<pcm->length; i++)
    {
      dac_idle->left[i]  = sample2DA(pcm->samples[0][i]);
      dac_idle->right[i] = sample2DA(pcm->samples[1][i]);
    }
  dac_idle->ptr = 0;
  dac_idle->sample_rate = pcm->samplerate;
  /* Must be last - this enables/triggers the timer's buffer swap.  */
  dac_idle->length = pcm->length;

  bucket(B_printf);
#if 0
  /* Enable this to see each frame as it's output.  */
  iprintf("mp3: out %d s %d c, %d Hz\n",
	  pcm->length, pcm->channels, pcm->samplerate);
#endif
#if ENABLE_BUCKETS
  for (i=1; i<= B_printf; i++)
    iprintf("%d %6d\t", i, buckets[i]);
  iprintf("\n");
#endif

  bucket(B_output_p);
}

/*----------------------------------------------------------------------*/

/* Up to you to handle errors however you want.  */

static
enum mad_flow error(void *data,
		    struct mad_stream *stream,
		    struct mad_frame *frame)
{
  Buffer *buffer = (Buffer *)data;

  iprintf("decoding error 0x%04x (%s)\n",
	  stream->error, mad_stream_errorstr(stream));

  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

  return MAD_FLOW_CONTINUE;
}

/*----------------------------------------------------------------------*/

/* This is a sample MP3 stored in ROM.  */
extern unsigned char sample_mp3[];
extern unsigned char sample_mp3_end[];

main()
{
  int d0 = 0, d1 = 0;
  int i;
  struct mad_decoder decoder;
  int result;
  Buffer buffer;

  osboard_init ();

  tty_puts("OSBoard MP3 tests\n");

  /* Timer - we use MTU2 for the output sample timer.  If you choose
     another timer, adjust the code in the ISR as well, as the timer
     is reprogrammed on the fly as sample rates change. */
  _vectors[VECT(MTU2,TGIA2)] = mtu2_handler;
  IEN(MTU2,TGIA2) = 1;
  IPR(MTU2,TGIA2) = 5;

  MTU2.TCR.BYTE = 0x20; /* TGRA, PCLK/1 = 48 MHz */
  MTU2.TIER.BYTE = 0x01; /* interrupt on TGRA */
  MTU2.TGRA = 1088; /* should be 44100 KHz */
  MTUA.TSTR.BIT.CST2 = 1; /* start timer */

  /* DAC0 is left */
  /* DAC1 is right */

  /* Only needed for our ROM sample */
  buffer.start = sample_mp3;
  buffer.length = sample_mp3_end - sample_mp3;

  /* Always needed, once.  */
  buffer.ibuf = ibuf;
  buffer.ilength = 0;
  buffer.isize = sizeof (ibuf);
  memset (dac_buf, 0, sizeof(dac_buf));
  dac_active = dac_buf + 0;
  dac_idle = dac_buf + 1;

  iprintf("libMAD version %s\n", mad_version);
  iprintf("libMAD %s\n", mad_copyright);
  iprintf("libMAD build: %s\n", mad_build);

  /* This is for our read-from-file demo.  */
  if (!microsd_card_ready())
    {
      iprintf("waiting for sdcard...\n");
      while (!microsd_card_ready())
	;
    }

  if (open_file ("sample.mp3") < 0)
    {
      iprintf("no sample.mp3 ?\n");
      while (1);
    }

  /* Chose INPUT_FILE or INPUT_ROM, and set the BUFFER fields
     accordingly. */
  mad_decoder_init(&decoder, &buffer,
		   input_file,
		   0 /* header */,
		   0 /* filter */,
		   output,
		   error,
		   0 /* message */);
  result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

  /* Done.  Exit, return, whatever.  */
  while (1)
    {
      asm("wait");
    }
}
